
#pragma once
#ifndef __COUNTEREXPRESSION_H__
#define __COUNTEREXPRESSION_H__

#include "boolexpression.h"

namespace cexp {

    class Environment;
    class CExpVisitor;

    typedef expr::BoolExpression<Environment, CExpVisitor>        CounterExpression;
    typedef expr::BoolExpressionConst<Environment, CExpVisitor>   CExpConst;
    typedef expr::BoolExpressionAtom<Environment, CExpVisitor>    CExpAtom;
    typedef expr::BoolExpressionBinOp<Environment, CExpVisitor>   CExpBinOp;
    typedef expr::BoolExpressionAnd<Environment, CExpVisitor>     CExpAnd;
    typedef expr::BoolExpressionOr<Environment, CExpVisitor>      CExpOr;
    typedef expr::BoolExpressionImplies<Environment, CExpVisitor> CExpImplies;
    typedef expr::BoolExpressionNot<Environment, CExpVisitor>     CExpNot;
    class CExpCompare;

    typedef boost::shared_ptr<Environment> environment_ptr;

    typedef expr::Ptrs<Environment, CExpVisitor>::boolexpression_ptr  counterexp_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_const_ptr   cexp_const_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_atom_ptr    cexp_atom_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_and_ptr     cexp_and_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_or_ptr      cexp_or_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_implies_ptr cexp_implies_ptr;
    typedef expr::Ptrs<Environment, CExpVisitor>::boolexp_not_ptr     cexp_not_ptr;

    typedef boost::shared_ptr<CExpCompare> cexp_compare_ptr;

    // Dummy environment since we'll never evaluate
    class Environment { };

    class CExpVisitor {
        public:
            virtual void visit(CExpConst& b) = 0;
            virtual void visit(CExpAnd& b) = 0;
            virtual void visit(CExpOr& b) = 0;
            virtual void visit(CExpImplies& b) = 0;
            virtual void visit(CExpNot& b) = 0;
            virtual void visit(CExpCompare& b) = 0;
    };


    enum CompOp { EQ, NEQ, GT, LT, GTE, LTE };

    class CExpCompare : public CExpAtom {
        std::string counter_name;
        CompOp      op;
        int         value;

        public:
            CExpCompare() { }

            CExpCompare(std::string const& new_counter_name,
                        CompOp new_op,
                        int new_value) 
                : counter_name(new_counter_name),
                  op(new_op),
                  value(new_value) {}
            ~CExpCompare() {}

            void set_counter_name(std::string const& new_counter_name) {
                counter_name = new_counter_name;
            }

            void set_op(CompOp new_op) {
                op = new_op;
            }

            void set_value(int new_value) {
                value = new_value;
            }

            std::string const& get_counter_name() const {
                return counter_name;
            }

            CompOp get_op() const {
                return op;
            }

            int get_value() const {
                return value;
            }

            virtual void to_stream(std::ostream& output) const;

            virtual bool evaluate(Environment const& env) const {
                return true;
            }

            virtual counterexp_ptr make_shared_ptr() {
                cexp_compare_ptr p(this);
                return p;
            }

            virtual void accept(CExpVisitor& v) { v.visit(*this); }

            virtual counterexp_ptr normal_form(bool negative);
    };



}

#endif
